#ifndef GHS_CONTROL_STRUCTS_H__
#define GHS_CONTROL_STRUCTS_H__

#include "stdbool.h"
#include "MderFloat.h"

#define OPCODE_LENGTH  1                        /**< Length of opcode inside PO Measurement packet. */
#define HANDLE_LENGTH  2                        /**< Length of handle inside PO Measurement packet. */
#define MAX_CHAR_LEN   (BLE_L2CAP_MTU_DEF - OPCODE_LENGTH - HANDLE_LENGTH)  /**< Maximum size of a transmitted PO Measurement. */

#define COMMAND_GET_PAIRING_REQUIRED 0x0009     // Command sent by PHG to ask if pairing is required by the PHD.
                                                // This is the FIRST command the PHG must send after service discovery
                                                // and enabling the characteristics. Enabling the characteristics will
                                                // never require security.
                                                // If the PHG and PHD are already paired OR the PHG knows that this PHD
                                                // is unpairable from a previous connection this command is not sent.
                                                // If the PHG and PHD are already paired, encryption starts.
                                                // The PHD requests pairing and the PHG does not invoke pairing, the
                                                // PHD will send an error on the next command and disconnect.
                                                
#define COMMAND_GET_SYS_INFO 0x000A             // This command gets the system info. Since it is static for the lifetime
                                                // of the PHD, if the PHG already has this data it need not ask for it again.
                                                // This command must come before the requesting of any measurement data.
                                                
#define COMMAND_GET_CONFIG_INFO 0x000B          // This command asks for config data. Support by the PHD is optional. The
                                                // details of the response have not been specified but it will have something
                                                // like a list of measurements supported by the PHD, such as a list of MDC
                                                // Type codes. This command must come before the requesting of any measurement data.
                                                
#define COMMAND_GET_CURRENT_TIME 0x000C         // This command gets the current time info. The PHG shall send this command every
                                                // connection. This command must come before the requesting of any measurement data.
                                                
#define COMMAND_SET_CURRENT_TIME 0x000D         // If the PHD supports setting the time the PHG shall set the time if the time has
                                                // not been set. If the time is set, the PHD shall adjust the time stamps of any 
                                                // unsent measurements accordingly to the new time line and all subsequent measurements
                                                // will be on this new time line. Only absolute and base offset times can be set.
                                                // The PHG will never have to handle date-time-adjustments. This command must
                                                // come before the requesting of any measurement data.
                                                // If the value of the time sync is TIME_SYNC_NONE the time has not be set.
                                                
#define COMMAND_GET_NUMBER_OF_STORED_RECORDS 0x000E // This command asks the PHD for the number of stored records. A record is
                                                    // a measurement group. A PHD must always respond to this command whether or
                                                    // not it stores data.
                                                    
#define COMMAND_GET_ALL_STORED_RECORDS 0x000F       // This command requests all stored records. The PHG may skip this command if
                                                    // the number of stored records is 0 OR the PHG does not want the data.
                                                    
#define COMMAND_DELETE_ALL_STORED_RECORDS 0x0010    // The PHG may request deletion of the stored data. The PHD may not support this
                                                    // command.
                                                    
#define COMMAND_SEND_LIVE_DATA 0x0011               // The PHG sends this command to request live data. Once this command is sent, the
                                                    // PHG shall not request any of the previous commands. At the moment no command has
                                                    // been defined for stopping the live data.
#define COMMAND_PROPRIETARY 0xFFFF

#define SYSINFO_FLAGS_REG_STATUS 1
#define SYSINFO_FLAGS_SERIAL_NO 2
#define SYSINFO_FLAGS_FIRMWARE 4
#define SYSINFO_FLAGS_SOFTWARE 8
#define SYSINFO_FLAGS_HARDWARE 16
#define SYSINFO_FLAGS_AVAS 32

#define HEADER_FLAGS_TIMESTAMP 1
#define HEADER_FLAGS_SUPP_TYPES 2;
#define HEADER_FLAGS_SRC_HNDL_REF 4;
#define HEADER_FLAGS_DURATION 8
#define HEADER_FLAGS_PERSONID 16
#define HEADER_FLAGS_SETTING 32
#define HEADER_FLAGS_HAS_AVAS 64
#define HEADER_FLAGS_CONT_FIRST 128
#define HEADER_FLAGS_CONT_FOLLOWS 256
#define HEADER_FLAGS_CONT_RECORD_DONE 512

#define MSMT_FLAGS_MSMT_TYPE_MASK 7
#define MSMT_FLAGS_SIMPLE_NUMERIC 0
#define MSMT_FLAGS_COMPOUND_NUMERIC 1
#define MSMT_FLAGS_CODED_ENUM 2
#define MSMT_FLAGS_16BITS_ENUM 3
#define MSMT_FLAGS_32BITS_ENUM 4
#define MSMT_FLAGS_RTSA 5
#define MSMT_FLAGS_STRING_ENUM 6
#define MSMT_FLAGS_EXT_OBJ 7
#define MSMT_FLAGS_SUPP_TYPES 16
#define MSMT_FLAGS_SRC_HNDL_REF 32
#define MSMT_FLAGS_DURATION 64
#define MSMT_FLAGS_AVAS 128
#define MSMT_FLAGS_SFLOAT_VAL 256

#define TIME_FLAGS_ABS_TIME 1
#define TIME_FLAGS_BO_TIME 2
#define TIME_FLAGS_REL_TIME 4
#define TIME_FLAGS_SET_TIME 8
#define TIME_FLAGS_TIME_SET 16
#define TIME_FLAGS_TIME_FAULT 32

#define PACKET_TYPE_NORMAL 0
#define PACKET_TYPE_CONT_FIRST 1
#define PACKET_TYPE_CONT_FOLLOWS 2

#define GHSCP_COMMAND_DONE 0
#define GHSCP_COMMAND_RECORD_DONE 1
#define GHSCP_COMMAND_UNSUPPORTED 2
#define GHSCP_COMMAND_UNKNOWN 3
#define GHSCP_COMMAND_ERROR 4

#define BLUETOOTH_SPECIALIZATION_NOT_FOUND -1
#define BLUETOOTH_MEASUREMENTS_SENT 0
#define BLUETOOTH_INITIALIZATION_FAILED 1
#define BLUETOOTH_ADDRESS_SETTING_FAILED 2
#define BLUETOOTH_DEVICE_NAME_SETTING_FAILED 3
#define BLUETOOTH_DIS_CREATION_FAILED 4
#define BLUETOOTH_CTS_CREATION_FAILED 5
#define BLUETOOTH_BATTERY_SERVICE_CREATION_FAILED 6
#define BLUETOOTH_ADVERTISMENT_CREATION_FAILED 7
#define BLUETOOTH_SPECIALIZATION_SERVICE_FAILED 8
#define BLUETOOTH_CHARACTERISTIC_CREATION_FAILED 9
#define BLUETOOTH_DATE_TIME_CHAR_FAILED 10
#define BLUETOOTH_FEATURE_CHAR_FAILED 11
#define BLUETOOTH_ADVERTISING_START_FAILED 12
#define BLUETOOTH_SPECIALIZATION_DISCONNECTED 13
#define BLUETOOTH_SPECIALIZATION_ABORTED 14
#define BLUETOOTH_NOT_CONFIGURED 20
#define BLUETOOTH_OK 100

#define BTLE_SCCD_GHS_SERVICE 0xF990
#define BTLE_SCCD_GHS_CP_CHAR 0xF991
#define BTLE_SCCD_GHS_RESPONSE_CHAR 0xF992


#define CONT_NONE 0
#define CONT_RECORD_SEND 1
#define CONT_RECORD_DONE 2

typedef struct 
{
    unsigned short chunks_outstanding;  // chunks sent and not 'acked' (mainly for notifications)
    unsigned short offset;              // Needed to handle fragmentation. 
    unsigned char* data;                // The data buffer being indicated/notified
    unsigned short data_length;         // the total length of the data buffer being indicated 
    unsigned short handle;              // the handle of the characteristic to make the indications/notifications on
    unsigned short current_command;     // the current command being handled
    unsigned short chunk_size;          // maximum length of each indication/notification
    unsigned short number_of_groups;    // how many records to send
    unsigned short continuous_stage;    // when true, this is part of a continuous sequence
} s_global_send;

typedef struct
{
    unsigned long attrId;
    unsigned short length;
    unsigned char* value;
} s_Avas;

typedef struct
{
    unsigned short specialization;      // term code from partition 8
    unsigned short version;             // specialization version
} s_Specialization;

typedef struct
{
    unsigned short command;
    unsigned short flagsRegStatus;
    unsigned short flagsSerialNo;
    unsigned short flagsFirmware;
    unsigned short flagsSoftware;
    unsigned short flagsHardware;
    unsigned short flagsAvas;
    unsigned short length;
    unsigned char systemId[8];
    unsigned short currentSpecializationCount;
    unsigned short numberOfSpecializations;
    s_Specialization* specializations;
    char* manufacturer;
    char* modelNumber;
    unsigned short regulationStatus;
    char* serialNo;
    char* firmware;
    char* software;
    char* hardware;
    unsigned short currentAvaCount;
    unsigned short numberOfAvas;
    s_Avas** avas;

} s_SystemInfo;

typedef struct
{
    unsigned short timeFlagsAbsTime;    // If equal to TIME_FLAGS_ABS_TIME PHD uses absolute time
    unsigned short timeFlagsBoTime;     // If equal to TIME_FLAGS_BO_TIME PHD uses base offset time
    unsigned short timeFlagsRelTime;    // If equal to TIME_FLAGS_REL_TIME PHD uses relative time
    unsigned short timeFlagsSetTime;    // If equal to TIME_FLAGS_SET_TIME PHD supports setting the time
    unsigned short timeFlagsTimeSet;    // If equal to TIME_FLAGS_TIME_SET PHD has had its time set
    unsigned short timeFlagsTimeFault;  // If equal to TIME_FLAGS_TIME_FAULT PHD has had a time fault and stored data times are unreliable
    unsigned long timeResolution;       // The resolution of the clock in ?
    unsigned short timeSync;            // The term code from partition 8 giving the time sync

} s_TimeInfo;

typedef struct
{
    char* deviceName;
    unsigned char bluetoothAddress[6];
    s_SystemInfo* systemInfo;
    s_TimeInfo* currentTime;
    bool invokeServiceChanged;
}s_PhdInfo;


typedef struct
{
    s_MderFloat value;
    unsigned short units;
} s_SimpNumeric;

typedef struct
{
    unsigned long subType;              // the subtype 32 nomenclature code
    s_MderFloat subValue;               // the sub value

} s_Compound;

typedef struct
{
    unsigned short numberOfComponents;  // the number of sub components in the compound
    s_Compound** value;
    unsigned short units;
} s_CompoundNumeric;

typedef struct
{
    unsigned long code;                 // the measurement as a 32 bit nomenclature code
} s_CodedEnum;

typedef struct
{
    unsigned short length;
    unsigned long bitsEnum;             // the integer carrying the bits
    unsigned long stateEvent;           // whether the bit is a state (set) or event (cleared)
    unsigned long supportEvent;         // whether the bit is supported (set) or not (cleared)
} s_Bits32Enum;

typedef struct
{
    unsigned short length;
    unsigned short bitsEnum;             // the integer carrying the bits
    unsigned short stateEvent;           // whether the bit is a state (set) or event (cleared)
    unsigned short supportEvent;         // whether the bit is supported (set) or not (cleared)
} s_Bits16Enum;

typedef struct
{
    s_MderFloat scaleFactor;            // part of the y = mx + b. This is the 'm' value. y= original quantity, x = scaled value (FLOAT)
    s_MderFloat offset;                 // part of the y = mx + b. This is the 'b' value (FLOAT)
    s_MderFloat period;                 // the length in seconds (FLOAT)
    unsigned char sampleSize;           // Size of each sample in bits
    unsigned short numberOfSamples;     // The number of samples
    unsigned char* samples;             // the array of samples
    unsigned short units;               // the units
} s_Rtsa;

typedef struct
{
    unsigned long type;             // The 32-bit nomenclature code giving the measurement type
    unsigned short flagsMsmtType;           // Set to one of the following below:
                                            // Set to MSMT_FLAGS_SIMPLE_NUMERIC if a simple numeric
                                            // Set to MSMT_FLAGS_COMPOUND_NUMERIC if a compound numeric
                                            // Set to MSMT_FLAGS_CODED_ENUM if a coded measurement
                                            // Set to MSMT_FLAGS_16BITS_ENUM if a 16-bit BITs measurement
                                            // Set to MSMT_FLAGS_32BITS_ENUM if a 32-bit BITs measurement
                                            // Set to MSMT_FLAGS_RTSA if an RTSA
                                            // Set to MSMT_FLAGS_STRING_ENUM if a string measurement
    unsigned short flagsSuppTypes;          // Set to MSMT_FLAGS_SUPP_TYPES if measurement has supplemental types
    unsigned short flagsSrcHndlRef;         // Set to MSMT_FLAGS_SRC_HNDL_REF if measurement has a soruce handle reference
    unsigned short flagsDuration;           // Set to MSMT_FLAGS_DURATION if measurement has a duration
    unsigned short flagsAvas;               // Set to MSMT_FLAGS_AVAs if measurement has AVA structs
    unsigned short flagsSfloat;             // Set to MSMT_FLAGS_SFLOAT_VAL if numeric measurement value uses SFLOATs vs FLOATs
    unsigned short id;                      // The identifier of this measurement instance
    s_SimpNumeric* simpleNumeric;           // Simple Numeric entry
    s_CompoundNumeric* compoundNumeric;     // Compound Numeric entry
    s_CodedEnum* codedEnum;                 // Coded Enumeration entry
    s_Bits16Enum* bits16Enum;               // Bits 16 Enumeration entry
    s_Bits32Enum* bits32Enum;               // Bits 32 Enumeration entry
    s_Rtsa* rtsa;                           // RTSA entry
    unsigned short numberOfSuppTypes;
    unsigned long* suppTypes;               // The list of supplemental types if there are supplemental types.
                                            //     Only valid if the flagsSuppTypes = MSMT_FLAGS_SUPP_TYPES
    unsigned short numberOfSrcHandleRefs;
    unsigned short* srcHandleRefs;          // Points to the measurement instance this measurement references if there is such a reference
                                            //     Only valid if the flagsSrcHndlRef = MSMT_FLAGS_SRC_HNDL_REF
    s_MderFloat *duration;                  // The duration in seconds as an Mder FLOAT if there is a duration
                                            //     Only valid if the flagsDuration = MSMT_FLAGS_DURATION
    unsigned short currentAvaCount;
    unsigned short numberOfAvas;
    s_Avas** avas;                          // The additional attributes as AVA structs relevant for this measurement
                                            //     Only valid if the flagsAvas = MSMT_FLAGS_AVAS
} s_GhsMsmt;

typedef struct
{
    unsigned short command;
    unsigned short flagsTimeStamp;
    unsigned short flagsSuppTypes;          // Set to HEADER_FLAGS_SUPP_TYPES if all measurements have a common supplemental types
    unsigned short flagsSrcHndlRef;         // Set to HEADER_FLAGS_SRC_HNDL_REF if all measurements have a common source handle reference
    unsigned short flagsDuration;           // Set to HEADER_FLAGS_DURATION if all measurements have a common duration
    unsigned short flagsPersonId;           // Set to HEADER_FLAGS_PERSONID if there is a person id
    unsigned short flagsSetting;            // Set to HEADER_FLAGS_SETTING if the measurements are settings
    unsigned short flagsAvas;               // Set to HEADER_FLAGS_HAS_AVAS if all measurements have common AVAs
    unsigned short flagsFirstCont;          // Set to HEADER_FLAGS_CONT_FIRST if the group is first in a sequence of optimized groups
    unsigned short flagsFollowsCont;        // Set to HEADER_FLAGS_CONT_FOLLOWS if the group is one of the optimized groups after the first
    unsigned short flagsRecordCont;         // Set to HEADER_FLAGS_CONT_RECORD_DONE if the group is an optinzed record complete PDU for notification
                                            //     record complete use instead of indications on the CP. Not sure if this is worth the effort of
                                            //     supporting.
    unsigned char numberOfMsmts;            // Number of measurements in the group
    unsigned char currentMsmtCount;         // For adding GhsMsmts into group

    unsigned short numberOfSuppTypes;       // The number of supplemental types
    unsigned long* suppTypes;               // The list of supplemental types if there are supplemental types.
                                            //     Only valid if the flagsSuppTypes = HEADER_FLAGS_SUPP_TYPES and the numberOfSuppTypes > 0
    unsigned short numberOfSrcHandleRefs;   // The number of supplemental types
    unsigned short* srcHandleRefs;          // Points to the measurement instance this all measurement references in this group references if there is such a reference
                                            //     Only valid if the flagsSrcHndlRef = HEADER_FLAGS_SRC_HNDL_REF
    s_MderFloat *duration;                  // The duration in seconds as an Mder FLOAT if there is a duration
                                            //     Only valid if the flagsDuration = HEADER_FLAGS_DURATION
    unsigned short personId;                // only valid id flagsPersonId = HEADER_FLAGS_PERSONID
    unsigned short currentAvaCount;
    unsigned short numberOfAvas;
    s_Avas** avas;                          // The additional attributes as AVA structs relevant for this measurement
                                            //     Only valid if the flagsAvas = MSMT_FLAGS_AVAS
    unsigned char groupId;                  // Used only if one of the HEADER_FLAGS_CONT_* are set, but always present. This group id is the same for the same
                                            // group no matter how many times the group is instantiated. It is therefore different than the measurement id which
                                            // is unique for every measurement instance. The group id is more like the 20601 handle. The measurement id is more
                                            // like the FHIR resource logical id
} s_MsmtHeader;

typedef struct
{
    s_MsmtHeader* header;
    s_GhsMsmt** ghsMsmts;
}s_MsmtGroup;

typedef struct
{
    unsigned short dataLength;
    unsigned short flags_index;
    short currentTime_index;
    short timeSync_index;
    unsigned char *timeInfoBuf;
}s_TimeInfoData;

typedef struct
{
    unsigned short dataLength;
    unsigned char *systemInfoBuf;
}s_SystemInfoData;

/**
 * The idea here is that any real agent will create a few templates for the measurements the device sends. When measurement
 * values are updated from the sensor, the agent will insert these values in the template. For example the following is the
 * byte sequence sent over the airwaves for a pulse ox with SpO2, PR, and Pulse Quality using SPOT modality:
     0F 00 01 00 3E 00 [A8 56 E7 00 00 00 00 00] 02 03        Header
     B8 4B 02 00 0D 00 81 08 [01 00] 20 02 [60 00] 01 3C 4C 02 00  SpO2
     1A 48 02 00 0D 00 81 08 [02 00] A0 0A [2E 00] 01 3C 4C 02 00  Pulse Rate
     30 4B 02 00 08 00 01 08 [03 00] 20 02 [44 E2]               Pulsatile Quality
 * Only the bytes in square brackets will change when one gets an update from the sensor; all the other values remain the same.
 * So the agent creates this template, stores it in ROM, and then when active, copies it into RAM and updates the terms in square
 * brackets as it gets new data and sends it. This is easy to do for a singular use case, but difficult to do generically.
 * Generically one would populate the structures, update the fields in the structures, and recreate the data packet each time
 * the sensor updates. Here we are trying to generically create the data template.
 *
 * The s_MsmtGroupData struct will have the template for the measurement group created with empty fields that need to be populated
 * with sensor info. But to do that one needs the indicies to the location of the data field to be populated. That's what the
 * s_GhsMsmtIndex struct has. There is one of these for each measurement in the group. The entries in the s_MsmtGroupData.ghsMsmtIndex[]
 * have the s_GhsMsmtIndex struct for each measurement in the order they were added to the group.
 * s_GhsMsmtIndex.id_index gives the start index to the id for that measurement
 * s_GhsMsmtIndex.value_index gives the index to the start of the value for that measurement. This works easily except for compounds where
 * one has multiple values. The number of values is given by s_GhsMsmtIndex.numberOfValues which is 1 for everything but compounds.
 * From the value_index one has to enter the first, skip the 4 bytes of the sub-type code, and enter the next, etc. For the numeric
 * and compound entries one also needs to know whether it is a 2-byte SFLOAT or 4-byte FLOAT.
 */

typedef struct
{
    unsigned short metricType;      // What type of metric this is. One of:
                                    //  MSMT_FLAGS_SIMPLE_NUMERIC
                                    //  MSMT_FLAGS_COMPOUND_NUMERIC
                                    //  MSMT_FLAGS_CODED_ENUM
                                    //  MSMT_FLAGS_16BITS_ENUM
                                    //  MSMT_FLAGS_32BITS_ENUM
                                    //  MSMT_FLAGS_RTSA
                                    //  MSMT_FLAGS_STRING_ENUM
    unsigned short id_index;        // The index in the MsmtGroup data[] buffer for the measurement id
    unsigned short numberOfValues;  // The number of measurement values. It is always 1 except for compounds
                                    //   where there may be more than one.
    unsigned short value_index;     // The index in the MsmtGroup data[] buffer for the first measurement value
                                    // The size of the measurement value depends upon the metric type
                                    //  MSMT_FLAGS_SIMPLE_NUMERIC 2 bytes or 4 bytes for SFLOAT or FLOAT
                                    //  MSMT_FLAGS_COMPOUND_NUMERIC 2 bytes or 4 bytes each entry for SFLOAT or FLOAT
                                    //      Only the first index is given. After the first is done one needs to skip 
                                    //      four bytes for the 4-byte sub type MDC code and then insert the next value
                                    //      and so forth until all values are filled.
                                    //  MSMT_FLAGS_CODED_ENUM 4 bytes
                                    //  MSMT_FLAGS_16BITS_ENUM 2 bytes
                                    //  MSMT_FLAGS_32BITS_ENUM 4 bytes
                                    //  MSMT_FLAGS_RTSA variable given by the number of samples
                                    //  MSMT_FLAGS_STRING_ENUM do a memcpy of the string
    bool isSfloat;                  //  If a simple numeric or compound numeric, indicates use of SFLOAT or FLOAT
}s_GhsMsmtIndex;

typedef struct
{
    unsigned short dataLength;          // total length of the data buffer
    unsigned short currentGhsMsmtCount; // The structure that was used to generate the data buffer
    s_GhsMsmtIndex **ghsMsmtIndex;      // the array of support info for each measurement entry
    unsigned char *data;
}s_MsmtGroupData;                       // Support information for using the measurement group data buffer for this measurement group

#define HAS_REL_TIMESTAMP 1
#define HAS_ABS_TIMESTAMP 0
#define USES_NUMERIC 1
#define USES_COMPOUND 1
#define USES_CODED 1
#define USES_BITS16 1
#define USES_BITS32 1
#define USES_RTSA 1
#define USES_SYSTEM_AVAS 1
#define USES_HEADER_AVAS 1
#define USES_MSMT_AVAS 1

typedef struct
{
    unsigned char century;
    unsigned char year;
    unsigned char month;
    unsigned char day;
    unsigned char hour;
    unsigned char min;
    unsigned char sec;
    unsigned char hundredth;
} s_Abs;

#define BP_CUFF 1
#define PULSE_OX 0
#define NUMBER_OF_STORED_MSMTS 30

#if (BP_CUFF == 1)
typedef struct
{
    bool hasAbsTimeStamp;
    s_Abs timeStamp;
    bool hasRelTimeStamp;
    unsigned char relTimeStamp[8];
    unsigned short systolic;
    unsigned short diastolic;
    unsigned short mean;
    unsigned short pulseRate;
}s_MsmtDataBp;
#endif

#if (PULSE_OX == 1)
typedef struct
{
    bool hasAbsTimeStamp;
    s_Abs timeStamp;
    bool hasRelTimeStamp;
    bool isContinuous;
    unsigned char relTimeStamp[8];
    unsigned short spo2;
    unsigned short pulseRate;
    unsigned short pulseQuality;
}s_MsmtDataPo;
#endif


#endif
